__title__ = 'MindsDB'
__package_name__ = 'mindsdb'
__version__ = '23.6.4.0'
__description__ = "MindsDB server, provides server capabilities to mindsdb native python library"
__email__ = "jorge@mindsdb.com"
__author__ = 'MindsDB Inc'
__github__ = 'https://github.com/mindsdb/mindsdb'
__pypi__ = 'https://pypi.org/project/mindsdb/'
__license__ = 'GPL-3.0'
__copyright__ = 'Copyright 2018- mindsdb'
