__title__ = 'MindsDB DynamoDB handler'
__package_name__ = 'mindsdb_dynamodb_handler'
__version__ = '0.0.1'
__description__ = "MindsDB handler for DynamoDB"
__author__ = 'Minura Punchihewa'
__github__ = 'https://github.com/mindsdb/mindsdb'
__pypi__ = 'https://pypi.org/project/mindsdb/'
__license__ = 'GPL-3.0'
__copyright__ = 'Copyright 2022- mindsdb'
