__title__ = "MindsDB PayPal handler"
__package_name__ = "mindsdb_paypal_handler"
__version__ = "0.0.1"
__description__ = "MindsDB handler for PayPal"
__author__ = "Minura Punchihewa"
__github__ = "https://github.com/mindsdb/mindsdb"
__pypi__ = "https://pypi.org/project/mindsdb/"
__license__ = "GPL-3.0"
__copyright__ = "Copyright 2023 - mindsdb"
